(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_sched.mli 381 2011-03-16 17:06:40Z gerd $ *)

(** Scheduler *)

open Mapred_tasks

type plan_config

val configure_plan : ?keep_temp_files:bool ->
                     Mapred_def.mapred_job_config ->
                     Mapred_config.mapred_config -> plan_config
  (** [configure_plan jc conf]

      Parameters:
      - [keep_temp_files]: if true, temporary files created during the
        map/reduce execution are not immediately deleted
   *)

type plan
  (** A plan contains:
      - tasks
      - dependencies between tasks
      - whether a task is done or not done
   *)

val create_plan : Plasma_client.plasma_cluster -> 
                  plan_config -> 
                    plan
  (** Creates a plan. The plan is empty. *)

val bigblock_size : plan -> int
  (** The effective size of bigblocks. This is the value passed to
      [configure_plan] (via jc) rounded up to the next multiple of blocks.
   *)

val add_inputs : plan -> unit
  (** Add the input files, and create map tasks. This involves the analysis
      of the block layout of the input files (expensive operation).
   *)

val add_map_output : plan -> int -> (file_tag * file) list -> Unix.inet_addr ->
                       unit
  (** Add these files as output of this map/emap task identified by this ID

      The IP addr points to the machine that executed the map or emap task
      (which is also the likely storage for the files)
   *)

val plan_complete : plan -> bool
  (** Respond whether everything is planned *)

val complete_inputs : plan -> unit
  (** Declare that no more inputs will be added. This triggers
      the rest of the graph construction.
   *)

val executable_tasks : plan -> task list
  (** Returns all tasks that are executable. This list can be quite long!
      The order of this list suggests an optimal order of execution
      (from a "static" point of view)
   *)

val hosts : plan -> (string * Unix.inet_addr) list
  (** returns all configured task server hosts *)

val mark_as_finished : plan -> task -> unit
  (** Marks this task as finished *)

val mark_as_started : plan -> task -> Unix.inet_addr -> int -> bool -> unit
  (** Marks this task as started, together with the node where it is running,
      and the runtime task ID and the misallocation flag
   *)

val remove_marks : plan -> task -> unit
  (** Revokes [mark_as_started] or [mark_as_finished] *)

val task_depends_on_list : plan -> task -> task list
  (** Returns the list of tasks a given task depends on *)

val plan_finished : plan -> bool
  (** Whether everything is done *)

val n_running : plan -> int
val n_finished : plan -> int
val n_total : plan -> int
  (** Stats of tasks *)

val avg_running : plan -> float
  (** Average number of running tasks *)

val cluster : plan -> Plasma_client.plasma_cluster
  (** Just return the cluster passed in *)

val print_plan : plan -> unit
  (** Debug printing to stdout *)

val generate_svg : plan -> string
  (** Return the graph as SVG (once plan complete) *)

val task_stats : plan -> task -> int * int
  (** Returns:
      - number of input files
      - number of non-local input files

      Raise [Not_found] if the task has never been started
   *)
